#!ruby -Ks
# -*- mode: ruby; encoding: sjis -*-
# Last updated: <2014/12/16 02:38:12 +0900>
#

require 'dxruby'
require 'nkf'
require 'benchmark'

# BDFtHg`pNX
#
# @version 0.0.2
#
# @note iso8859  jisx0208 ݂̂ɑΉ
#
# @example
#    require 'dxruby'
#    require_relative 'bdffont'
#    bdf = BDFFont.new("mplus_j12r_jisx0208.bdf")
#    bdf.add("mplus_f12r_iso8859.bdf")
#    Window.loop do
#      x, y = 10, 10
#      Window.draw_bdffont(x, y, "This is a BDF font", bdf)
#      y += bdf.height
#      Window.draw_bdffont(x, y, "The quick brown fox jumps over the lazy dog.", bdf)
#      y += bdf.height
#      Window.draw_bdffont(x, y, "1234567890 ! = - +", bdf)
#      y += bdf.height
#      Window.draw_bdffont(x, y, "JAPANESE MESSAGE OK", bdf)
#    end
#
class BDFFont

  # @private
  # BDFtHg1̃f[^i[NX
  class BDFchar

    # @return [Object] ImageIuWFNg
    attr_reader :image

    # @return [Integer] ̕`ʒu x ItZbg
    attr_reader :ofsx

    # @return [Integer] ̕`ʒu y ItZbg
    attr_reader :ofsy

    # @return [Integer] 1`悵ɐi߂xW
    attr_reader :dwidth

    # @return [Integer] 摜
    attr_reader :w

    # @return [Integer] 摜c
    attr_reader :h

    # @return [Integer] R[h
    attr_reader :startchar

    attr_reader :dheight
    attr_reader :encoding
    attr_reader :swidth
    attr_reader :sheight

    # RXgN^
    #
    # @param [Integer] startchar R[h
    # @param [Array] col FBa,r,g,b ׂzŎw
    def initialize(startchar, col)
      @startchar = startchar
      @col = col
      @encoding = 0
      @swidth, @sheight = 0, 0
      @dwidth, @dheight = 0, 0
      @w, @h = 0, 0
      @ofsx, @ofsy = 0, 0
      @image = nil
      @bitpat = []
      @bitmap_read = false
    end

    # ^ꂽ͂ĕf[^Ƃēo^
    #
    # @param [String] l ׂ͂
    # @return [Boolean] trueȂ1ǂݎBfalseȂǂݎ葱s
    def parse(l)
      if @bitmap_read
        if l.index("ENDCHAR") == 0
          # rbg}bvf[^`IBimage𐶐
          @image = make_image(@w, @h, @bitpat, @col)
          @bitmap_read = false
          return true
        else
          # rbg}bvz𗭂߂
          @bitpat.push(l)
        end
      else
        p = l.split(' ')
        case p[0]
        when "ENCODING"
          @encoding = p[1].to_i
        when "SWIDTH"
          @swidth, @sheight = p[1].to_i, p[2].to_i
        when "DWIDTH"
          @dwidth, @dheight = p[1].to_i, p[2].to_i
        when "BBX"
          @w, @h = p[1].to_i, p[2].to_i
          @ofsx, @ofsy = p[3].to_i, p[4].to_i
        when "BITMAP"
          @bitmap_read = true
        end
      end
      return false
    end

    # 16i̔z񂩂Image𐶐ĕԂ
    #
    # @param [Integer] w 
    # @param [Integer] h c
    # @param [Array] dt 16i񂪓z
    # @param [Array] col FBa,r,g,b ׂzŎw
    # @return [Object] ImageIuWFNg
    def make_image(w, h, dt, col)
      img = Image.new(w, h)
      msk = 1 << (dt[0].length * 4 - 1)
      v = 0
      h.times do |y|
        v = dt[y].hex
        w.times do |x|
          img[x, y] = col if (v & (msk >> x)) != 0
        end
      end
      return img
    end
  end

  # @private
  # BDFtHg1t@C̃f[^i[NX
  class BDFdata

    # @return [Hash] BDFtHg1pNX𕡐i[nbV
    attr_reader :bdf_dic

    # @return [Integer] |CgTCY
    attr_reader :pt_size

    # @return [Integer] DPI
    attr_reader :xdpi

    # @return [Integer] cDPI
    attr_reader :ydpi

    # @return [Integer] oEfBO{bNX̉
    attr_reader :width

    # @return [Integer] oEfBO{bNX̏c
    attr_reader :height

    # @return [Integer] oEfBO{bNXxItZbg
    attr_reader :ofsx

    # @return [Integer] oEfBO{bNXyItZbg
    attr_reader :ofsy

    # @return [Integer] sNZTCY
    attr_reader :pixel_size

    # @return [Integer] R[hB:iso8859, :jisx0208, :iso10646 蓾
    attr_reader :charset

    # RXgN^
    #
    # @param [String] bdffilename BDFtHgt@C
    # @param [Array] col FBa,r,g,b ׂzŎw
    def initialize(bdffilename, col = [255, 255, 255, 255])
      @bdf_dic = {}
      @pt_size = 0
      @xdpi, @ydpi = 0, 0
      @ofsx, @ofsy = 0, 0
      @width, @height = 0, 0
      @pixel_size = 0
      @charset = nil
      @cahrs = 0

      # t@Cǂݍ
      f = open(bdffilename)
      filedata = f.read
      f.close

      bdfc = nil
      charcode = -1
      filedata.each_line do |l|
        # 1s
        l.chomp!
        if charcode >= 0
          # ̃rbg}bvf[^ǂݎ蒆
          if bdfc.parse(l)
            @bdf_dic[charcode] = bdfc
            charcode = -1
          end
        else
          p = l.split(' ')
          case p[0]
          when "STARTCHAR"
            # ̃rbg}bvf[^`Jn
            charcode = p[1].hex
            bdfc = BDFchar.new(charcode, col)
          when "SIZE"
            # |CgAdpi
            @pt_size, @xdpi, @ydpi = p[1].to_i, p[2].to_i, p[3].to_i
          when "FONTBOUNDINGBOX"
            # oEfBO{bNX
            @width, @height = p[1].to_i, p[2].to_i
            @ofsx, @ofsy = p[3].to_i, p[4].to_i
          when "PIXEL_SIZE"
            # sNZTCY
            @pixel_size = p[1].to_i
          when "CHARSET_REGISTRY"
            # R[h
            p[1].gsub!('"', '')
            case p[1].downcase
            when "iso8859"
              @charset = :iso8859
            when "jisx0208.1990"
              @charset = :jisx0208
            when "iso10646"
              @charset = :iso10646
            end
          when "CHARS"
            # S
            @chars = p[1].to_i
          when "ENDFONT"
            # Sf[^ǂݎI
            break
          end
        end
      end
    end

  end

  # @return [Integer] oEfBO{bNX̏c(ől)
  attr_reader :height

  # RXgN^
  #
  # @param [String] bdffilename BDFtHgt@C
  # @param [Array] col FBa,r,g,b ׂzŎw
  #
  def initialize(bdffilename, col = [255, 255, 255, 255])
    @bdffiles = []
    bdfdt = BDFdata.new(bdffilename, col)
    @bdffiles.push(bdfdt)
    @width = bdfdt.width
    @height = bdfdt.height
  end

  # BDFt@CǉB
  #
  # BDFtHg(jisx0208)ŃNX𐶐A
  # pBDFtHg(iso8859)ꏏɊ܂߂鎞ɎgB
  #
  # @param [String] bdffilename BDFtHgt@C
  # @param [Array] col FBa,r,g,b ׂzŎw
  #
  def add(bdffilename, col = [255, 255, 255, 255])
    bdfdt = BDFdata.new(bdffilename, col)
    @bdffiles.push(bdfdt)
    @width = bdfdt.width if @width <= bdfdt.width
    @height = bdfdt.height if @height <= bdfdt.height
  end

  # zƂĎw肵ۂɁABDFtHg1p̃NXԂ
  #
  # @param [Integer] index zY
  # @return [Object] BDFtHg1̃NX
  # @return [nil] nilȂo^͌Ȃ
  #
  def [](index)
    @bdffiles.each do |d|
      code = index
      case d.charset
      when :jisx0208
        code = self.sjis2jisx0208(code.ord)
      when :iso10646
        # BSJISiso10646ւ̕ϊ@Ȃc
        code = NKF.nkf('-w -S --cp932', code).ord
      else
        code = code.ord
      end
      if d.bdf_dic.has_key?(code)
        return d.bdf_dic[code]
      end
    end
    return nil
  end

  # @private
  # SJISR[hJISX0208R[hɕϊĕԂ
  # @param [Integer] code SJISR[h
  # @return [Integer] JISX0208R[h
  def sjis2jisx0208(code)
    code -= 0x4000 if code >= 0x0e000
    ub = (code & 0x0ff00) >> 8
    code = ((ub - 0x081) << (1+8)) + (code & 0x00ff)
    code -= 1 if (code & 0x00ff) >= 0x80
    if (code & 0x00ff) >= 0x009e
      code += (0x0100 - 0x009e)
    else
      code -= 0x0040
    end
    code += 0x2121
    return code
  end

end

# WindowNXɒǉBDFtHg`p\bh
module Window

  # BDFtHg`
  # @param [Number] x `xW
  # @param [Number] y `yW
  # @param [String] str `敶
  # @param [Object] bdf BDFtHgNX
  # @param [Integer] z `D揇
  def self.draw_bdffont(x, y, str, bdf, z = 0)
    str.split('').each do |code|
      b = bdf[code]
      unless b == nil
        Window.draw(x + b.ofsx, y - b.ofsy, b.image, z)
        x += b.dwidth
      end
    end
  end
end

# ImageNXɒǉBDFtHg`p\bh
class Image

  # BDFtHg`
  # @param [Number] x `xW
  # @param [Number] y `yW
  # @param [String] str `敶
  # @param [Object] bdf BDFtHgNX
  def draw_bdffont(x, y, str, bdf)
    str.split('').each do |code|
      b = bdf[code]
      unless b == nil
        self.draw(x + b.ofsx, y - b.ofsy, b.image)
        x += b.dwidth
      end
    end
  end
end

if $0 == __FILE__
  #eXgR[h

  lst = [
    "!\"\#$%&'()*+,-./0123456789:;<=>?@",
    "A B C D EFGHIJKLMNOPQRSTUVWXYZ[\]^_`",
    "abcdefghijklmnopqrstuvwxyz{|}~",
    "",
    "VB]̒",
    "e̖SCŏ̎瑹΂肵ĂB",
    "qɓΊpBɞΗB"
  ]

  col = [255, 128, 255, 128]

  # BDFtHgNX𐶐

  bdf = nil

  Benchmark.bm do |x|
    x.report {
      case 2
      when 0
        # ----------------------------------------
        # mplus 10 dot jisx0208 + iso8859

        bdf = BDFFont.new("mplus_j10r_jisx0208.bdf", col)
        bdf.add("mplus_f10r_iso8859.bdf", col)
      when 1
        # ----------------------------------------
        # mplus 12 dot jisx0208 + iso8859

        bdf = BDFFont.new("mplus_j12r_jisx0208.bdf", col)
        bdf.add("mplus_f12r_iso8859.bdf", col)
      when 2
        # ----------------------------------------
        # milk 16x16 dot jisx0208 + iso8859

        bdf = BDFFont.new("milkjf_k16_jisx0208.bdf", col)
        bdf.add("milkjf_8x16_iso8859.bdf", col)
      when 3
        # ----------------------------------------
        # mplus 10 dot iso10646 (Ή)

        bdf = BDFFont.new("umplus_j10r_iso10646.bdf", col)
      end
    }
  end

  # 摜ɕ`悵Ă݂
  image = Image.new(640, 240)
  x, y = 0, 0
  lst.each do |s|
    image.draw_bdffont(x, y, s, bdf)
    y += bdf.height
  end

  Window.bgcolor = [96, 96, 96]

  Window.loop do
    break if Input.keyPush?(K_ESCAPE)

    # ʂɕ`悵Ă݂
    x, y = 0, 0
    lst.each do |s|
      Window.draw_bdffont(x, y, s, bdf)
      y += bdf.height
    end

    Window.draw(0, 240, image)

    bar = Window.height * Window.getLoad / 100.0
    Window.drawBoxFill(Window.width - 2, 0, Window.width, bar, C_RED)
  end

end
